/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#include "puredatawrapper.h"

#include <wx/mimetype.h>
#include <wx/utils.h>
#include <wx/intl.h>
#include <wx/regex.h>

#include "exceptions.h"

enum {	// id for sockets
	SERVER_ID = 100, SOCKET_ID
};

BEGIN_EVENT_TABLE(CPureDataWrapper, wxEvtHandler)    
    EVT_END_PROCESS(wxID_ANY, CPureDataWrapper::OnProcessTerm)
//	EVT_SOCKET(SERVER_ID,  CPureDataWrapper::OnServerEvent)
	EVT_SOCKET(SOCKET_ID,  CPureDataWrapper::OnSocketEvent)
END_EVENT_TABLE()

CPureDataWrapper::CPureDataWrapper() 
: m_process(this)
 , m_pid(0)
 , m_pdConnection(NULL)
 , m_status(CPureDataWrapper::NOT_RUNNING)
 , m_parserStatus(CPureDataWrapper::IGNORE_INPUT)
 , m_entry(false)
 , m_whichAPI(-1)
 , m_debugGUIMode(false)
 , m_logStream(NULL)
 , m_ownConsole(false)
{

}

CPureDataWrapper* CPureDataWrapper::g_single= NULL;

CPureDataWrapper* CPureDataWrapper::getInstance()
{
	//	static CPureDataWrapper single;

	// Although this is the best way to implement the singleton pattern
	// because it ensures that the destructor is always called, here doesn't
	// works because this class depends on other "static" classes and the destruction
	// order is not guaranted. So we allocate in the heap.

	if (g_single== NULL)
		g_single= new CPureDataWrapper();
	return g_single;
}

void CPureDataWrapper::destroyInstance()
{
	if (g_single!= NULL)
	{
		delete g_single;
		g_single= NULL;
	}
}

CPureDataWrapper::~CPureDataWrapper(void)
{
	StopPD ();
	DisableLog ();
}

void CPureDataWrapper::OnProcessTerm(wxProcessEvent& event)
{
	// Free resources
	m_pid= 0;
	m_pdVersion.Clear();
	m_whichAPI= -1;
	m_apiList.clear();
	EPdWrapperStatus status= m_status;
	m_status= CPureDataWrapper::NOT_RUNNING;
	event.Skip(false);

	if (status!= CPureDataWrapper::WAIT_INCOMING_CONNECTION && status!= CPureDataWrapper::WAIT_TERMINATE)
		throw RunTimeException (_("PdWrapper: Pure Data died unexpectedly."));
}

void CPureDataWrapper::ParseInput (const char* buffer, wxUint32 size)
{	
	//
	// TODO: filter pdtk_post messages to avoid messing with expected messages
	//

	static wxRegEx initRegEx (_T(" *pdtk_pd_startup *{"));
	static wxRegEx whichAPIRegEx (_T("[[:space:]]*set pd_whichapi[[:space:]]*"));
	static wxRegEx openPatchRegEx (_T("[[:space:]]*pdtk_canvas_new[[:space:]]+[.period.][[:alnum:]]+[[:space:]]"));
	static wxRegEx closeOkRegEx (_T("[[:space:]]*destroy *[.period.][[:alnum:]]*"));
	static wxRegEx closeErrorRegEx (_T("[[:space:]]*pdtk_post[[:space:]]*{.*:[[:space:]]*no such object"));
	static wxRegEx waitDSPRegEx (_T("pdtk_pd_dsp"));
	static wxRegEx intRegEx (_T("-?[[:digit:]]"));
	static wxRegEx apiListRegEx (_T("{[[:space:]]*{.*}[[:space:]]*}"));
	static wxRegEx apiListItemRegEx (_T("{[^{}]+}"));
	static wxRegEx audioPropIn1RegEx (_T("global *audio_indevlist;"));
	static wxRegEx audioPropIn2RegEx (_T("lappend audio_indevlist"));
	static wxRegEx audioPropOut1RegEx (_T("global *audio_outdevlist;"));
	static wxRegEx audioPropOut2RegEx (_T("lappend audio_outdevlist"));
	static wxRegEx audioPropValuesRegEx (_T("pdtk_audio_dialog *[.period.][[:alnum:]]+[[:space:]]"));
		
	assert (initRegEx.IsValid());
	assert (whichAPIRegEx.IsValid());
	assert (openPatchRegEx.IsValid());
	assert (closeOkRegEx.IsValid());
	assert (closeErrorRegEx.IsValid());
	assert (waitDSPRegEx.IsValid());
	assert (intRegEx.IsValid());
	assert (apiListRegEx.IsValid());
	assert (apiListItemRegEx.IsValid());
	assert (audioPropIn1RegEx.IsValid());
	assert (audioPropIn2RegEx.IsValid());
	assert (audioPropOut1RegEx.IsValid());
	assert (audioPropOut2RegEx.IsValid());
	assert (audioPropValuesRegEx.IsValid());

	m_parser.NextBuffer (buffer, (unsigned short) size);

	unsigned short lineSize= 0;
	const char* line= m_parser.getLine(&lineSize);
	while (line) {
		wxString wline(line, wxConvLocal, lineSize);
		if (m_status== CPureDataWrapper::INIT1) {
			// Try to parse version
			if (initRegEx.Matches(wline)) {
				size_t start, len;
				if (initRegEx.GetMatch(&start, &len)) {
					m_pdVersion= wline.Mid (len + start);
					m_status= CPureDataWrapper::INIT2;
				}
			}
		} else if (m_status== CPureDataWrapper::INIT2) {
			// Parse audio API devices
            if (apiListRegEx.Matches(wline)) {
				wxString strApiList= apiListRegEx.GetMatch(wline);
				while (apiListItemRegEx.Matches(strApiList)) {
					TApiDesc apiDesc;
					size_t start, len;
					if (apiListItemRegEx.GetMatch(&start, &len)) {
                        wxString strApi= apiListItemRegEx.GetMatch(strApiList);
						strApi.Replace(_T("{"), _T(""));
						strApi.Replace(_T("}"), _T(""));
						apiDesc.apiName= strApi.AfterFirst(_T('"')).BeforeLast(_T('"'));
						if (strApi.AfterLast(_T('"')).ToLong(&apiDesc.apiNum)) {
							m_apiList.push_back (apiDesc);
							m_status= CPureDataWrapper::INIT3;
						}
						
						strApiList= strApiList.Mid(start + len);
					}					
				}
			}
		} else if (m_status== CPureDataWrapper::INIT3) {
			// Try to parse which API
			if (whichAPIRegEx.Matches(wline))
				if (intRegEx.Matches(wline))
					if (intRegEx.GetMatch(wline).ToLong(&m_whichAPI))
						m_status= CPureDataWrapper::RUNNING;				
		} else if (m_status== CPureDataWrapper::RUNNING) {
			switch (m_parserStatus)	{
			case CPureDataWrapper::IGNORE_INPUT:
				break;
			case CPureDataWrapper::WAIT_ACK:	// Change state when data received
				if (wline.Length()> 0) m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
				break;
			case CPureDataWrapper::WAIT_OPEN_PATCH:
				if (openPatchRegEx.Matches(wline)) {
					m_tmpExchange= openPatchRegEx.GetMatch(wline);
					m_tmpExchange.Trim();
					m_tmpExchange.Trim(false);
					m_tmpExchange= m_tmpExchange.AfterFirst(_T(' '));
					m_parserStatus= CPureDataWrapper::IGNORE_INPUT;				
				}
				break;
			case CPureDataWrapper::WAIT_CLOSE_PATCH:
				if (closeOkRegEx.Matches(wline)) {
                    m_tmpError= false;
					m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
				}
				else if (closeErrorRegEx.Matches(wline)) {
					m_tmpError= true;
					m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
				}
				break;		
			case CPureDataWrapper::AUDIO_PROP_IN1:
				if (audioPropIn1RegEx.Matches(wline)) {
					m_audioProperties.m_iDevList.clear();
					m_parserStatus= CPureDataWrapper::AUDIO_PROP_IN2;
				}
				break;
			case CPureDataWrapper::AUDIO_PROP_IN2:
				if (audioPropIn2RegEx.Matches(wline)) {
					size_t start, len;
					if (audioPropIn2RegEx.GetMatch(&start, &len)) {
                        wxString str= wline.Mid (start+len);
						str.Replace(_T("{"), _T(""));
						str.Replace(_T("}"), _T(""));
						str.Trim();
						str.Trim(false);
						m_audioProperties.m_iDevList.push_back (str);
					}
				}
				else {
					// Try to parse output devices
					m_parserStatus= CPureDataWrapper::AUDIO_PROP_OUT1;
					continue;
				}					
				break;
			case CPureDataWrapper::AUDIO_PROP_OUT1:
				if (audioPropOut1RegEx.Matches(wline)) {
					m_audioProperties.m_oDevList.clear();
					m_parserStatus= CPureDataWrapper::AUDIO_PROP_OUT2;
				}
				break;
			case CPureDataWrapper::AUDIO_PROP_OUT2:
				if (audioPropOut2RegEx.Matches(wline)) {
					size_t start, len;
					if (audioPropOut2RegEx.GetMatch(&start, &len)) {
                        wxString str= wline.Mid (start+len);
						str.Replace(_T("{"), _T(""));
						str.Replace(_T("}"), _T(""));
						str.Trim();
						str.Trim(false);
						m_audioProperties.m_oDevList.push_back (str);
					}
				}
				else {
					// Try to parse output devices
					m_parserStatus= CPureDataWrapper::AUDIO_PROP_VALUES;
					continue;
				}					
				break;
			case CPureDataWrapper::AUDIO_PROP_VALUES:
				if (audioPropValuesRegEx.Matches(wline)) {
					// Dialog name
					m_tmpExchange.Clear();
					m_tmpExchange= audioPropValuesRegEx.GetMatch(wline);
					m_tmpExchange.Trim();
					m_tmpExchange.Trim(false);
					m_tmpExchange= m_tmpExchange.AfterLast(_T(' '));

					// Get value list
					size_t start, len;
					audioPropValuesRegEx.GetMatch(&start, &len);
					int numRead= sscanf(line + start + len, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
						&m_audioProperties.m_iDev1, &m_audioProperties.m_iDev2, &m_audioProperties.m_iDev3, 
						&m_audioProperties.m_iDev4, &m_audioProperties.m_iDev1Channels, &m_audioProperties.m_iDev2Channels, 
						&m_audioProperties.m_iDev3Channels, &m_audioProperties.m_iDev4Channels, &m_audioProperties.m_oDev1, 
						&m_audioProperties.m_oDev2, &m_audioProperties.m_oDev3, &m_audioProperties.m_oDev4,
						&m_audioProperties.m_oDev1Channels, &m_audioProperties.m_oDev2Channels, 
						&m_audioProperties.m_oDev3Channels, &m_audioProperties.m_oDev4Channels,
                        &m_audioProperties.m_sampleRate, &m_audioProperties.m_msDelay, &m_audioProperties.m_canMulti,
						&m_audioProperties.m_flongForm);
					if (numRead!= 20) m_tmpError= true;
					else m_tmpError= false;
					m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
				}
				break;
			case CPureDataWrapper::WAIT_WHICH_API:
				if (whichAPIRegEx.Matches(wline))
					if (intRegEx.Matches(wline))
						if (intRegEx.GetMatch(wline).ToLong(&m_whichAPI))
							m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
				break;
			}			
		} else {
			// 
		}
		line= m_parser.getLine(&lineSize); 
	}
}

void CPureDataWrapper::OnSocketEvent(wxSocketEvent& event)
{
	event.Skip(false);

	switch (event.GetSocketEvent())
	{
	case wxSOCKET_LOST:
		assert (m_pdConnection);
		m_pdConnection->Close();
		m_pdConnection->Discard();
		m_pdConnection->Destroy();
		m_pdConnection= NULL;
		if (m_status!= CPureDataWrapper::WAIT_TERMINATE) StopPD();
		break;
	case wxSOCKET_INPUT:
		{
		char readBuff[2048];
		assert (m_pdConnection);
		m_pdConnection->Read (readBuff, 2048);
		if (m_pdConnection->Error())
			throw RunTimeException (_("PdWrapper: Error reading from socket."));

		ParseInput (readBuff, m_pdConnection->LastCount());
		
		// log
		if (m_logStream) {
			fwrite (readBuff, 1, m_pdConnection->LastCount(), m_logStream);
			fflush (m_logStream);
		}

		}
		break;
	default:
		throw RunTimeException (_("PdWrapper: Unexpected socket event."));
	}
}

#define YIELD()	{				\
		if (wxThread::IsMain())	\
            wxSafeYield();		\
        else					\
            wxThread::Yield();	\
	}

bool CPureDataWrapper::WaitWhileParserStatusIs (EParserStatus e, int time_cs)
{
	int waitCount= 0;
	while (waitCount< time_cs && m_parserStatus== e) {
		wxMilliSleep(100);
		YIELD();
		waitCount++;
	}
	return (m_parserStatus!= e);
}

bool CPureDataWrapper::WaitWhileParserStatusIsNot (EParserStatus e, int time_cs)
{
	int waitCount= 0;
	while (waitCount< time_cs && m_parserStatus!= e) {
		wxMilliSleep(100);
		YIELD();
		waitCount++;
	}
	return (m_parserStatus== e);
}

class SetReset {
public:
	SetReset (bool* var) {	m_val= var; *m_val= true;	}
	~SetReset()	{ *m_val= false; }
private:
	bool* m_val;
};

wxString CPureDataWrapper::CorrectFilePath (const wxString& fp) const
{
	wxString retval(fp);
	retval.Replace(_T("\\"), _T("/"));
	retval.Replace(_T(" "), _T("\\ "));
	return retval;
}

void CPureDataWrapper::SendMessageToPD (const wxString& msg)
{
	const wxCharBuffer msg_ansi= msg.mb_str(wxConvLocal);
	m_pdConnection->Write (msg_ansi, (wxUint32) msg.Length());	
}

void CPureDataWrapper::LaunchPD (const wxString& params)
{
	// Get pd binary
	wxMimeTypesManager mimeManager;		
	wxFileType* ft= mimeManager.GetFileTypeFromExtension(_T("pd"));
	if (ft== NULL)
		throw InitializationException (_("PdWrapper: .pd extension not associated with PureData\nIs Pure Data (PD) installed?"));

	wxString pdBinary;
	if (!ft->GetOpenCommand(&pdBinary, wxFileType::MessageParameters(wxEmptyString)))
	{
		delete ft;
		throw InitializationException (_("PdWrapper: Cannot get the command to start PureData\nIs Pure Data (PD) installed?"));
	}
	delete ft;

	// Set command line parameters
	pdBinary.Replace (_T("\"\""), _T(""));
	pdBinary.Trim();
	pdBinary.Append (params);

	// Run pd
	m_pid= wxExecute(pdBinary, wxEXEC_ASYNC, &m_process);
	if (!m_pid)
		throw InitializationException (_("PdWrapper: Cannot run PureData\nIs Pure Data (PD) properly installed?"));
}

void CPureDataWrapper::StartPD ()
{
	if (m_debugGUIMode) return;

	if (m_status== CPureDataWrapper::RUNNING) return;
	wxSocketServer* serverSocket= NULL;

	try
	{
		// Code to avoid reentrant calls from the same thread
		if (m_entry) return;
		SetReset sr(&m_entry);
		assert (m_status== CPureDataWrapper::NOT_RUNNING);

		//
		// Create a listening socket
		//
		wxIPV4address bindAddress;		
		bool bindOk= false;
		if (!bindAddress.AnyAddress())
			throw InitializationException (_("PdWrapper: Error setting bind address."));
		// Look for a free port
		unsigned short service= 60000;
		while (!bindOk && service> 40000)
		{
			if (!bindAddress.Service(service))
				throw InitializationException (_("PdWrapper: Error setting bind port."));
			
			// Try to create 
			serverSocket= new wxSocketServer (bindAddress);
			if (serverSocket->IsOk()) bindOk= true;	// Bind successful
			else 
			{
				delete serverSocket;
				service--;
			}
		}

		if (!bindOk) 
			throw InitializationException (_("PdWrapper: Error creating server socket. Cannot bind to any port."));

		//
		// Launch pd
		//
		m_status= CPureDataWrapper::WAIT_INCOMING_CONNECTION;
		wxString param;
		param.Printf (_T(" -guiport %u"), service);
		LaunchPD (param);

		// 
		// Wait for an incomming connection
		//	
		assert (m_pdConnection== NULL);
		int waitCount= 0;
		// Wait for incoming connection (5 sec)
		while (waitCount< 50 && m_status== CPureDataWrapper::WAIT_INCOMING_CONNECTION) 
		{
			// Note that "WaitForAccept" will call wxYield so further events are processed while waiting
			if (serverSocket->WaitForAccept(0,100))
			{
				// Connection received
				//if (!serverSocket->AcceptWith(m_pdConnection))
				m_pdConnection= serverSocket->Accept(false);
				if (m_pdConnection== NULL)
					throw InitializationException (_("PdWrapper: Error while trying to stablish connection with Pure Data."));
				m_status= CPureDataWrapper::INIT1;			
			}
			waitCount++;
		}
		if (m_status== CPureDataWrapper::NOT_RUNNING)
			throw InitializationException (_("PdWrapper: Pure Data process died unexpectedly while waiting for incoming connection."));
		else if (waitCount>= 50)
			throw InitializationException (_("PdWrapper: Timeout while waiting for incoming connection."));
		//delete serverSocket;
		serverSocket->Destroy();
		serverSocket= NULL;

		//
		// Connect socket events
		//
		m_pdConnection->Notify (false);
		m_pdConnection->SetNotify (wxSOCKET_INPUT_FLAG | wxSOCKET_LOST_FLAG);
		m_pdConnection->SetEventHandler (*this, SOCKET_ID);		
		m_pdConnection->Notify (true);

		//
		// Read initialization strings
		//
		waitCount= 0;
		while (waitCount< 50 && m_status!= CPureDataWrapper::RUNNING)
		{
			wxMilliSleep(100);
			YIELD();
			waitCount++;
		}
		if (m_status!= CPureDataWrapper::RUNNING)
			throw InitializationException (_("PdWrapper: Unexpected error during waiting for PD initialization."));

		//
		// Send pd init message
		//
		m_parserStatus= CPureDataWrapper::WAIT_ACK;
		SendMessageToPD (_T("pd init ") + CorrectFilePath(wxGetCwd()) + _T(";"));		
		if (!WaitWhileParserStatusIs (CPureDataWrapper::WAIT_ACK))
			throw InitializationException (_("PdWrapper: Timeout while waiting for pd init acknowledgment."));

		// Make sure PD is ready
		wxMilliSleep(500);
	}
	catch(...)
	{
		// Free resources & rethrow
		if (serverSocket) serverSocket->Destroy();
		StopPD(); 		
		throw;
	}   
}

void CPureDataWrapper::StopPD ()
{
	// Code to avoid reentrant calls from the same thread
	if (m_entry) return;
	SetReset sr(&m_entry);

	if (m_status== CPureDataWrapper::NOT_RUNNING ||
		m_status== CPureDataWrapper::WAIT_TERMINATE) return;
	
	assert (m_status!= CPureDataWrapper::WAIT_INCOMING_CONNECTION);
	if (m_status== CPureDataWrapper::WAIT_INCOMING_CONNECTION)
		throw RunTimeException (_("PdWrapper: Inconsistent process status while stopping"));

	int waitCounter= 0;

	m_status= CPureDataWrapper::WAIT_TERMINATE;

	if (!m_debugGUIMode) {
		// Try to gently stop pd sending a "pd quit;" message
		if (m_pdConnection && m_pdConnection->IsConnected()) {
			static char* buff= "pd quit;";
			m_pdConnection->Write (buff, (wxUint32) strlen (buff));

			while (waitCounter< 50 && m_status!= CPureDataWrapper::NOT_RUNNING) {
				wxMilliSleep(100);
				YIELD();
				waitCounter++;
			}

			if (m_pdConnection) m_pdConnection->Destroy();
			m_pdConnection= NULL;
			if (m_status== CPureDataWrapper::NOT_RUNNING) return;
		}
	}

	// If the previous method did not work send a SIGTERM and wait again
	m_process.Kill (m_pid, wxSIGTERM);
	waitCounter= 0;
	while (waitCounter< 20 && m_status!= CPureDataWrapper::NOT_RUNNING)
	{
			wxMilliSleep(100);
			YIELD();
			waitCounter++;
	}
	if (m_status== CPureDataWrapper::NOT_RUNNING) return;

	// Finally kill the process
	m_process.Kill (m_pid, wxSIGKILL);
	waitCounter= 0;
	while (waitCounter< 50 && m_status!= CPureDataWrapper::NOT_RUNNING)
	{
			wxMilliSleep(100);
			YIELD();
			waitCounter++;
	}
	if (m_status== CPureDataWrapper::NOT_RUNNING) return;

	// The code should never reach here. If so we only assert because 
	// StopPD may have been called from the destructor
	assert (false);
}

// Check whether the call is not reentrant and PD is running, used by
// almost all public methods. Returns false if is not possible to run
// the method now (i.e. the caller should return)
bool CPureDataWrapper::CommonChecks ()
{
	// Code to avoid reentrant calls from the same thread
	if (m_entry) return false;
	SetReset sr(&m_entry);

	if (m_status!= CPureDataWrapper::RUNNING)
		throw RunTimeException (_("PdWrapper: PD not running"));

	return true;
}

wxString CPureDataWrapper::OpenPatch (const wxString& file)
{
	// Check if file exists
	if (!::wxFileExists(file))
		throw FileNotFoundException (_("PdWrapper: File not found: ") + file);	

	if (!m_debugGUIMode)
	{
		if (!CommonChecks ()) return wxString();
		
		m_parserStatus= CPureDataWrapper::WAIT_OPEN_PATCH;

		wxString name= wxFileNameFromPath(file);
		wxString path= wxPathOnly(file);
		SendMessageToPD (_T("pd open ") + name + _T(" ") + CorrectFilePath(path) + _T(";"));
		if (!WaitWhileParserStatusIs (CPureDataWrapper::WAIT_OPEN_PATCH)) {
			m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
			throw InitializationException (_("PdWrapper: Timeout opening patch."));
		}

		return m_tmpExchange;
	}
	else
	{
		// GUI Debug mode
		StopPD();	// Stop pd in case it was running

		LaunchPD (_T(" \"") + file + _T("\""));
		
		m_status= CPureDataWrapper::RUNNING;

		return _T("foo");
	}
}

void CPureDataWrapper::ClosePatch (const wxString& patchId)
{
	if (!m_debugGUIMode && !CommonChecks ()) return;

	if (!m_debugGUIMode) {
		m_parserStatus= CPureDataWrapper::WAIT_CLOSE_PATCH;
		m_tmpError= false;
		SendMessageToPD (patchId + _T(" menuclose 0 ;"));
		if (!WaitWhileParserStatusIs (CPureDataWrapper::WAIT_CLOSE_PATCH)) {
			m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
			throw RunTimeException (_("PdWrapper: Timeout closing patch."));
		}

		if (m_tmpError)
			throw NoSuchObjectException (_("PdWrapper: Closing patch. No such object: ") + patchId);
	} 
	else StopPD();	// Debug mode
	
}

void CPureDataWrapper::StartDSP ()
{
	if (m_debugGUIMode) return;
	if (!CommonChecks ()) return;

	// Don't wait for ACK. To avoid error when the DSP was already enabled
	SendMessageToPD (_T("pd dsp 1 ;"));	
}

void CPureDataWrapper::StopDSP ()
{
	if (m_debugGUIMode) return;
	if (!CommonChecks ()) return;

	SendMessageToPD (_T("pd dsp 0 ;"));	
}

void CPureDataWrapper::ManageAudioOptionsDialog (const wxString& msg)
{
	m_parserStatus= CPureDataWrapper::AUDIO_PROP_IN1;
	SendMessageToPD (msg);
	if (!WaitWhileParserStatusIsNot (CPureDataWrapper::IGNORE_INPUT)) {
		m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
		throw RunTimeException (_("PdWrapper: Timeout reading audio properties."));
	}

	bool errorFlag= m_tmpError;
	// Close audio dialogue	
	m_parserStatus= CPureDataWrapper::WAIT_CLOSE_PATCH;
	SendMessageToPD (m_tmpExchange + _T(" cancel ;"));
	if (!WaitWhileParserStatusIs (CPureDataWrapper::WAIT_CLOSE_PATCH)) {
		m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
		throw RunTimeException (_("PdWrapper: Timeout while closing audio properties dialogue."));
	}

	if (errorFlag) throw RunTimeException (_("PdWrapper: Unexpected error while parsing audio properties."));

	if (m_tmpError) throw RunTimeException (_("PdWrapper: Unexpected error while closing audio properties dialogue."));
}

CAudioProperties* CPureDataWrapper::getAudioProperties ()
{
	if (m_debugGUIMode || !CommonChecks ()) return NULL;
		
	ManageAudioOptionsDialog (_T("pd audio-properties ;"));

	return new CAudioProperties(m_audioProperties);
}


void CPureDataWrapper::setAudioProperties (const CAudioProperties& ap, bool savePreferences)
{
	if (m_debugGUIMode || !CommonChecks ()) return;

	wxString msg;
	msg.Printf (_T("pd audio-dialog %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d ;"),
		ap.m_iDev1, ap.m_iDev2, ap.m_iDev3, ap.m_iDev4, ap.m_iDev1Channels, ap.m_iDev2Channels, 
		ap.m_iDev3Channels, ap.m_iDev4Channels, ap.m_oDev1, ap.m_oDev2, ap.m_oDev3, ap.m_oDev4,
		ap.m_oDev1Channels, ap.m_oDev2Channels, ap.m_oDev3Channels, ap.m_oDev4Channels,
        ap.m_sampleRate, ap.m_msDelay, ap.m_flongForm);
	if (savePreferences) msg.Append (_T(" pd save-preferences ;"));
	m_parserStatus= CPureDataWrapper::WAIT_WHICH_API;
	SendMessageToPD (msg);
	if (!WaitWhileParserStatusIs (CPureDataWrapper::WAIT_WHICH_API)) {
		m_parserStatus= CPureDataWrapper::IGNORE_INPUT;
		throw RunTimeException (_("PdWrapper: Timeout while setting audio properties."));
	}
}

void CPureDataWrapper::setCurrentAPI(long apiNum)
{
	if (m_debugGUIMode || !CommonChecks ()) return;

	// Ensure apiNum is correct
	bool found= false;
	for (unsigned int i= 0; i< m_apiList.size(); i++) 
		if (m_apiList[i].apiNum== apiNum) {
			found= true;
			break;
		}

	if (!found) throw RunTimeException (_("PdWrapper: API id not available."));

	wxString msg;
	msg.Printf (_T("pd audio-setapi %d ;"), apiNum);
	ManageAudioOptionsDialog (msg);

	m_whichAPI= apiNum;
}

// Tries to automatically provide a suitable configuration over ASIO drivers
void CPureDataWrapper::SetIntelligentASIOConfig (unsigned long delay, bool savePreferences)
{
	if (m_debugGUIMode) return;

	// Search ASIO api number
	unsigned int i;
	for (i= 0; i< m_apiList.size(); i++) {
		wxString name= m_apiList[i].apiName;
		name.MakeLower();
		if (name.Contains(_T("asio"))) break;	// Found
	}

	if (i== m_apiList.size())
		throw RunTimeException (_("Can not detect ASIO drivers. Are they installed?"));

	// Switch to ASIO API
	setCurrentAPI(m_apiList[i].apiNum);
		
	// Read configuration
	CAudioProperties* ap= getAudioProperties();

	if (!ap) 
		throw RunTimeException (_T("Can not get audio properties."));

	try{
		//
		// Configure properties
		//

		// Input device
		unsigned int iDevNum;
		for (iDevNum= 0; iDevNum< ap->m_iDevList.size(); iDevNum++) 
			if (ap->m_iDevList[iDevNum].Lower().Contains(_T("asio"))) break;	// Found
		if (iDevNum== ap->m_iDevList.size())
			throw RunTimeException (_("Can not detect ASIO drivers. Are they installed?"));
		
		// Output device
		unsigned int oDevNum;
		for (oDevNum= 0; oDevNum< ap->m_oDevList.size(); oDevNum++) 
			if (ap->m_oDevList[oDevNum].Lower().Contains(_T("asio"))) break;	// Found
		if (oDevNum== ap->m_oDevList.size())
			throw RunTimeException (_("Can not detect ASIO drivers. Are they installed?"));
		
		// Set audio parameters
		ap->m_iDev1= iDevNum;
		ap->m_iDev2= ap->m_iDev3= ap->m_iDev4= 0;			
		ap->m_iDev1Channels= 2;		// Assumes two audio channels
		ap->m_iDev2Channels= ap->m_iDev3Channels= ap->m_iDev4Channels= 0;
		ap->m_oDev1= oDevNum;
		ap->m_oDev2= ap->m_oDev3= ap->m_oDev4= 0;
		ap->m_oDev1Channels=2;		// Assumes two audio channels
		ap->m_oDev2Channels= ap->m_oDev3Channels= ap->m_oDev4Channels= 0;
		ap->m_msDelay= delay;

		setAudioProperties (*ap, savePreferences);

		delete ap;
	}
	catch (...)
	{
		if (ap) delete ap;
		throw;
	}
}

// Provides a basic debug mode in which only one patch can be loaded
// at given time but with the full PD GUI open.
void CPureDataWrapper::setDebugGUIMode (bool enable)
{
	if (m_debugGUIMode== enable) return;

	// Stop pd in case it was running
	StopPD();
	
	m_debugGUIMode= enable;
}

void CPureDataWrapper::EnableLog (FILE *stream)
{
	if (stream== NULL) {
#ifdef __WXMSW__	
		AllocConsole();
		freopen("CONOUT$", "wb", stdout);	
#endif
		m_logStream= stdout;
		m_ownConsole= true;
	} else { 
		m_logStream= stream;
		m_ownConsole= false;
	}
}

void CPureDataWrapper::DisableLog ()
{
	if (m_ownConsole) FreeConsole();
	m_logStream= NULL;
	m_ownConsole= false;
}

///////////////////////////////////////

CPureDataSimpleController* CPureDataSimpleController::g_single= NULL;

CPureDataSimpleController* CPureDataSimpleController::getInstance()
{
	//	static CPureDataSimpleController single;

	// Although this is the best way to implement the singleton pattern
	// because it ensures that the destructor is always called, here doesn't
	// works because this class depends on other "static" classes and the destruction
	// order is not guaranted. So we allocate in the heap.

	if (g_single== NULL)
		g_single= new CPureDataSimpleController();
	return g_single;
}

void CPureDataSimpleController::destroyInstance()
{
	if (g_single!= NULL)
	{
		CPureDataWrapper::destroyInstance();
		delete g_single;
		g_single= NULL;
	}
}

wxString CPureDataSimpleController::OpenPatch (const wxString& file)
{
	wxString patchId;
	try {
		CPureDataWrapper::getInstance()->StartPD();
		CPureDataWrapper::getInstance()->StartDSP();
		patchId= CPureDataWrapper::getInstance()->OpenPatch(file);
		m_openPatchs.Add(patchId);
	}
	catch(...)
	{
		// If failed stop PD when necessary
		if (m_openPatchs.IsEmpty())
			CPureDataWrapper::getInstance()->StopPD();
		throw;
	}	

	return patchId;
}

void CPureDataSimpleController::ClosePatch (const wxString& patchId)
{
	int index= m_openPatchs.Index (patchId);
	if (index== wxNOT_FOUND)
		throw RunTimeException (_("PDController: ClosePatch: no such patchID: ") + patchId);

	m_openPatchs.RemoveAt (index);
	CPureDataWrapper::getInstance()->ClosePatch (patchId);

	if (m_openPatchs.IsEmpty())
			CPureDataWrapper::getInstance()->StopPD();
}

CPureDataSimpleController::CPureDataSimpleController()
{
}
	

////////////////////////////////////////////////////////////////////////////////////////
	/*
void CPureDataWrapper::RunPatch(void)
{

	if (m_pdPid!= 0) throw InitializationException (_("Patch already open"));

	// External patch
	wxMimeTypesManager mimeManager;
		
	wxFileType* ft= mimeManager.GetFileTypeFromExtension(_T("pd"));
	if (ft== NULL)
		throw InitializationException (_(".pd extension not associated with PureData\nIs Pure Data (PD) installed?"));

	wxString pdBinary;
	if (!ft->GetOpenCommand(&pdBinary, wxFileType::MessageParameters(wxEmptyString)))
	{
		delete ft;
		throw InitializationException (_("Cannot get the command to start PureData\nIs Pure Data (PD) installed?"));
	}
	delete ft;

	// TODO: ensure this works on Vista/7

	pdBinary.Replace (_T("\"\""), _T(""));
	pdBinary.Trim();
	//pdBinary.Replace (_T("\""), _T(""));
	pdBinary.Replace (_T(".exe"), _T(".com"));
		
	// Detect ASIO devices
	wxArrayString output, errors;
	wxString cmd(pdBinary);
	cmd.Append (_T(" -listdev -asio -stderr -nogui -version"));
	if (::wxExecute(cmd, output, errors, wxEXEC_SYNC )== -1)
		throw InitializationException (_("Cannot start PureData for automatic detection"));
	long asioIn= FindASIOInputDevice (errors);
	if (asioIn== -1) asioIn= FindASIOInputDevice (output);
	long asioOut= FindASIOOutputDevice (errors);
	if (asioOut== -1) asioOut= FindASIOOutputDevice (output);

	// TODO ASIO drivers automatical detection disabled until solved
#if 0
	if (asioIn== -1 || asioOut== -1)
		throw InitializationException (_("Cannot detect ASIO drivers. Perhaps haven't been installed."));
#endif

	// Run Pure Data
	cmd= pdBinary;
	if (m_noGui) cmd.Append (_T(" -nogui "));
	cmd.Append(	_T(" -asio -audiobuf ") + wxString::Format(_T("%d "), m_msAudioBuffer) + 
#if 0
				_T("-audioindev ") + wxString::Format(_T("%d "), asioIn) + 
				_T("-audiooutdev ") + wxString::Format(_T("%d "), asioOut) +
#endif
				_T("-open "));
	cmd.Append(_T("\"") + m_patchName + _T("\""));

	m_pdPid = ::wxExecute (cmd);
	if (m_pdPid== 0)
		throw InitializationException (_("Cannot start PureData"));
	
}


void CPureDataWrapper::KillPD(void)
{
	if (m_pdPid>0) ::wxKill (m_pdPid, wxSIGKILL);
	m_pdPid= 0;
}

long CPureDataWrapper::FindASIOInputDevice (const wxArrayString& pdOutput)
{
	return FindASIODevice (pdOutput, _T("audio input devices"));
}

long CPureDataWrapper::FindASIOOutputDevice(const wxArrayString& pdOutput)
{
	return FindASIODevice (pdOutput, _T("audio output devices"));
}

long CPureDataWrapper::FindASIODevice (const wxArrayString& pdOutput, const wxString& header)
{
	size_t i;
	long devNum= -1;

	// Find section
	for (i= 0; i< pdOutput.GetCount() && !pdOutput[i].StartsWith(header); i++);

	if (i== pdOutput.GetCount()) return -1;

	// Found, find ASIO entry
	for (++i; i< pdOutput.GetCount(); i++)
	{
		// Entries are supposed to start with a number
		wxString line= pdOutput[i].Trim(false).Lower();
		if (line.Contains(_T("asio")))		
		{
			size_t pos;
			for (pos= 0; pos< line.Length() && line[pos]>= wxT('0') && line[pos]<= wxT('9'); pos++);
			if (pos== 0) return -1;
			line[pos]=wxT('\0');
			if (!line.ToLong(&devNum)) return -1;
			break;
		}
	}
	if (i== pdOutput.GetCount()) return -1;
	
	return devNum;
}
*/
